@@ -38,11 +38,11 @@
  *
  * This code is released under the terms of the PostgreSQL License.
  *
- * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2010, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  * Portions taken from FreeBSD.
  *
- * $PostgreSQL: pgsql/src/bin/initdb/initdb.c,v 1.172.2.1 2009/11/14 15:39:45 mha Exp $
+ * $PostgreSQL: pgsql/src/bin/initdb/initdb.c,v 1.186 2010/02/26 02:01:15 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -120,10 +120,10 @@ static int	n_buffers = 50;
  * Warning messages for authentication methods
  */
 #define AUTHTRUST_WARNING \
-"# CAUTION: Configuring the system for local \"trust\" authentication allows\n" \
-"# any local user to connect as any PostgreSQL user, including the database\n" \
-"# superuser. If you do not trust all your local users, use another\n" \
-"# authentication method.\n"
+"# CAUTION: Configuring the system for local \"trust\" authentication\n" \
+"# allows any local user to connect as any PostgreSQL user, including\n" \
+"# the database superuser.  If you do not trust all your local users,\n" \
+"# use another authentication method.\n"
 static char *authwarning = NULL;
 
 /*
@@ -149,23 +149,22 @@ static char **replace_token(char **lines,
 #ifndef HAVE_UNIX_SOCKETS
 static char **filter_lines_with_token(char **lines, const char *token);
 #endif
-static char **readfile(char *path);
+static char **readfile(const char *path);
 static void writefile(char *path, char **lines);
 static FILE *popen_check(const char *command, const char *mode);
 static int	mkdir_p(char *path, mode_t omode);
 static void exit_nicely(void);
 static char *get_id(void);
 static char *get_encoding_id(char *encoding_name);
-static char *get_short_version(void);
 static int	check_data_dir(char *dir);
 static bool mkdatadir(const char *subdir);
 static void set_input(char **dest, char *filename);
 static void check_input(char *path);
-static void set_short_version(char *short_version, char *extrapath);
+static void write_version_file(char *extrapath);
 static void set_null_conf(void);
 static void test_config_settings(void);
 static void setup_config(void);
-static void bootstrap_template1(char *short_version);
+static void bootstrap_template1(void);
 static void setup_auth(void);
 static void get_set_pwd(void);
 static void setup_depend(void);
@@ -176,6 +175,7 @@ static void setup_dictionary(void);
 static void setup_privileges(void);
 static void set_info_version(void);
 static void setup_schema(void);
+static void load_plpgsql(void);
 static void vacuum_db(void);
 static void make_template0(void);
 static void make_postgres(void);
@@ -362,10 +362,10 @@ filter_lines_with_token(char **lines, const char *token)
  * get the lines from a text file
  */
 static char **
-readfile(char *path)
+readfile(const char *path)
 {
 	FILE	   *infile;
-	int			maxlength = 0,
+	int			maxlength = 1,
 				linelen = 0;
 	int			nlines = 0;
 	int			n;
@@ -395,19 +395,16 @@ readfile(char *path)
 	}
 
 	/* handle last line without a terminating newline (yuck) */
-
 	if (linelen)
 		nlines++;
 	if (linelen > maxlength)
 		maxlength = linelen;
 
 	/* set up the result and the line buffer */
-
-	result = (char **) pg_malloc((nlines + 2) * sizeof(char *));
-	buffer = (char *) pg_malloc(maxlength + 2);
+	result = (char **) pg_malloc((nlines + 1) * sizeof(char *));
+	buffer = (char *) pg_malloc(maxlength + 1);
 
 	/* now reprocess the file and store the lines */
-
 	rewind(infile);
 	n = 0;
 	while (fgets(buffer, maxlength + 1, infile) != NULL && n < nlines)
@@ -805,42 +802,6 @@ find_matching_ts_config(const char *lc_type)
 }
 
 
-/*
- * get short version of VERSION
- */
-static char *
-get_short_version(void)
-{
-	bool		gotdot = false;
-	int			end;
-	char	   *vr;
-
-	vr = xstrdup(PG_VERSION);
-
-	for (end = 0; vr[end] != '\0'; end++)
-	{
-		if (vr[end] == '.')
-		{
-			if (end == 0)
-				return NULL;
-			else if (gotdot)
-				break;
-			else
-				gotdot = true;
-		}
-		else if (vr[end] < '0' || vr[end] > '9')
-		{
-			/* gone past digits and dots */
-			break;
-		}
-	}
-	if (end == 0 || vr[end - 1] == '.' || !gotdot)
-		return NULL;
-
-	vr[end] = '\0';
-	return vr;
-}
-
 /*
  * make sure the directory either doesn't exist or is empty
  *
@@ -969,7 +930,7 @@ check_input(char *path)
  * if extrapath is not NULL
  */
 static void
-set_short_version(char *short_version, char *extrapath)
+write_version_file(char *extrapath)
 {
 	FILE	   *version_file;
 	char	   *path;
@@ -984,14 +945,14 @@ set_short_version(char *short_version, char *extrapath)
 		path = pg_malloc(strlen(pg_data) + strlen(extrapath) + 13);
 		sprintf(path, "%s/%s/PG_VERSION", pg_data, extrapath);
 	}
-	version_file = fopen(path, PG_BINARY_W);
-	if (version_file == NULL)
+
+	if ((version_file = fopen(path, PG_BINARY_W)) == NULL)
 	{
 		fprintf(stderr, _("%s: could not open file \"%s\" for writing: %s\n"),
 				progname, path, strerror(errno));
 		exit_nicely();
 	}
-	if (fprintf(version_file, "%s\n", short_version) < 0 ||
+	if (fprintf(version_file, "%s\n", PG_MAJORVERSION) < 0 ||
 		fclose(version_file))
 	{
 		fprintf(stderr, _("%s: could not write file \"%s\": %s\n"),
@@ -1249,14 +1210,14 @@ setup_config(void)
 		if (err != 0 ||
 			getaddrinfo("::1", NULL, &hints, &gai_result) != 0)
 			conflines = replace_token(conflines,
-									  "host    all         all         ::1",
-									  "#host    all         all         ::1");
+							   "host    all             all             ::1",
+							 "#host    all             all             ::1");
 	}
 #else							/* !HAVE_IPV6 */
 	/* If we didn't compile IPV6 support at all, always comment it out */
 	conflines = replace_token(conflines,
-							  "host    all         all         ::1",
-							  "#host    all         all         ::1");
+							  "host    all             all             ::1",
+							  "#host    all             all             ::1");
 #endif   /* HAVE_IPV6 */
 
 	/* Replace default authentication methods */
@@ -1294,7 +1255,7 @@ setup_config(void)
  * run the BKI script in bootstrap mode to create template1
  */
 static void
-bootstrap_template1(char *short_version)
+bootstrap_template1(void)
 {
 	PG_CMD_DECL;
 	char	  **line;
@@ -1314,7 +1275,7 @@ bootstrap_template1(char *short_version)
 	/* Check that bki file appears to be of the right version */
 
 	snprintf(headerline, sizeof(headerline), "# PostgreSQL %s\n",
-			 short_version);
+			 PG_MAJORVERSION);
 
 	if (strcmp(headerline, *bki_lines) != 0)
 	{
@@ -1396,20 +1357,6 @@ setup_auth(void)
 	PG_CMD_DECL;
 	const char **line;
 	static const char *pg_authid_setup[] = {
-		/*
-		 * Create triggers to ensure manual updates to shared catalogs will be
-		 * reflected into their "flat file" copies.
-		 */
-		"CREATE TRIGGER pg_sync_pg_database "
-		"  AFTER INSERT OR UPDATE OR DELETE ON pg_database "
-		"  FOR EACH STATEMENT EXECUTE PROCEDURE flatfile_update_trigger();\n",
-		"CREATE TRIGGER pg_sync_pg_authid "
-		"  AFTER INSERT OR UPDATE OR DELETE ON pg_authid "
-		"  FOR EACH STATEMENT EXECUTE PROCEDURE flatfile_update_trigger();\n",
-		"CREATE TRIGGER pg_sync_pg_auth_members "
-		"  AFTER INSERT OR UPDATE OR DELETE ON pg_auth_members "
-		"  FOR EACH STATEMENT EXECUTE PROCEDURE flatfile_update_trigger();\n",
-
 		/*
 		 * The authid table shouldn't be readable except through views, to
 		 * ensure passwords are not publicly visible.
@@ -1446,8 +1393,6 @@ get_set_pwd(void)
 
 	char	   *pwd1,
 			   *pwd2;
-	char		pwdpath[MAXPGPATH];
-	struct stat statbuf;
 
 	if (pwprompt)
 	{
@@ -1517,16 +1462,6 @@ get_set_pwd(void)
 	PG_CMD_CLOSE;
 
 	check_ok();
-
-	snprintf(pwdpath, sizeof(pwdpath), "%s/global/pg_auth", pg_data);
-	if (stat(pwdpath, &statbuf) != 0 || !S_ISREG(statbuf.st_mode))
-	{
-		fprintf(stderr,
-				_("%s: The password file was not generated. "
-				  "Please report this problem.\n"),
-				progname);
-		exit_nicely();
-	}
 }
 
 /*
@@ -1807,6 +1742,7 @@ setup_privileges(void)
 		"  WHERE relkind IN ('r', 'v', 'S') AND relacl IS NULL;\n",
 		"GRANT USAGE ON SCHEMA pg_catalog TO PUBLIC;\n",
 		"GRANT CREATE, USAGE ON SCHEMA public TO PUBLIC;\n",
+		"REVOKE ALL ON pg_largeobject FROM PUBLIC;\n",
 		NULL
 	};
 
@@ -1916,6 +1852,31 @@ setup_schema(void)
 	check_ok();
 }
 
+/*
+ * load PL/pgsql server-side language
+ */
+static void
+load_plpgsql(void)
+{
+	PG_CMD_DECL;
+
+	fputs(_("loading PL/pgSQL server-side language ... "), stdout);
+	fflush(stdout);
+
+	snprintf(cmd, sizeof(cmd),
+			 "\"%s\" %s template1 >%s",
+			 backend_exec, backend_options,
+			 DEVNULL);
+
+	PG_CMD_OPEN;
+
+	PG_CMD_PUTS("CREATE LANGUAGE plpgsql;\n");
+
+	PG_CMD_CLOSE;
+
+	check_ok();
+}
+
 /*
  * clean everything up in template1
  */
@@ -2217,21 +2178,14 @@ check_locale_encoding(const char *locale, int user_enc)
 
 	locale_enc = pg_get_encoding_from_locale(locale);
 
-	/* We allow selection of SQL_ASCII --- see notes in createdb() */
+	/* See notes in createdb() to understand these tests */
 	if (!(locale_enc == user_enc ||
 		  locale_enc == PG_SQL_ASCII ||
-		  user_enc == PG_SQL_ASCII
+		  locale_enc == -1 ||
 #ifdef WIN32
-
-	/*
-	 * On win32, if the encoding chosen is UTF8, all locales are OK (assuming
-	 * the actual locale name passed the checks above). This is because UTF8
-	 * is a pseudo-codepage, that we convert to UTF16 before doing any
-	 * operations on, and UTF16 supports all locales.
-	 */
-		  || user_enc == PG_UTF8
+		  user_enc == PG_UTF8 ||
 #endif
-		  ))
+		  user_enc == PG_SQL_ASCII))
 	{
 		fprintf(stderr, _("%s: encoding mismatch\n"), progname);
 		fprintf(stderr,
@@ -2386,7 +2340,7 @@ CreateRestrictedProcess(char *cmd, PROCESS_INFORMATION *processInfo)
 	}
 
 #ifndef __CYGWIN__
-    AddUserToTokenDacl(restrictedToken);
+	AddUserToTokenDacl(restrictedToken);
 #endif
 
 	if (!CreateProcessAsUser(restrictedToken,
@@ -2484,7 +2438,6 @@ main(int argc, char *argv[])
 				i,
 				ret;
 	int			option_index;
-	char	   *short_version;
 	char	   *effective_user;
 	char	   *pgdenv;			/* PGDATA value gotten from and sent to
 								 * environment */
@@ -2500,6 +2453,7 @@ main(int argc, char *argv[])
 		"pg_xlog",
 		"pg_xlog/archive_status",
 		"pg_clog",
+		"pg_notify",
 		"pg_subtrans",
 		"pg_twophase",
 		"pg_multixact/members",
@@ -2742,6 +2696,14 @@ main(int argc, char *argv[])
 	sprintf(pgdenv, "PGDATA=%s", pg_data);
 	putenv(pgdenv);
 
+	/*
+	 * Also ensure that TZ is set, so that we don't waste time identifying the
+	 * system timezone each of the many times we start a standalone backend.
+	 * It's okay to use a hard-wired value here because nothing done during
+	 * initdb cares about the timezone setting.
+	 */
+	putenv("TZ=GMT");
+
 	if ((ret = find_other_exec(argv[0], "postgres", PG_BACKEND_VERSIONSTR,
 							   backend_exec)) < 0)
 	{
@@ -2784,12 +2746,6 @@ main(int argc, char *argv[])
 
 	canonicalize_path(share_path);
 
-	if ((short_version = get_short_version()) == NULL)
-	{
-		fprintf(stderr, _("%s: could not determine valid short version string\n"), progname);
-		exit(1);
-	}
-
 	effective_user = get_id();
 	if (strlen(username) == 0)
 		username = effective_user;
@@ -2875,11 +2831,9 @@ main(int argc, char *argv[])
 
 		ctype_enc = pg_get_encoding_from_locale(lc_ctype);
 
-		if (ctype_enc == PG_SQL_ASCII &&
-			!(pg_strcasecmp(lc_ctype, "C") == 0 ||
-			  pg_strcasecmp(lc_ctype, "POSIX") == 0))
+		if (ctype_enc == -1)
 		{
-			/* Hmm, couldn't recognize the locale's codeset */
+			/* Couldn't recognize the locale's codeset */
 			fprintf(stderr, _("%s: could not find suitable encoding for locale %s\n"),
 					progname, lc_ctype);
 			fprintf(stderr, _("Rerun %s with the -E option.\n"), progname);
@@ -3121,7 +3075,7 @@ main(int argc, char *argv[])
 	check_ok();
 
 	/* Top level PG_VERSION is checked by bootstrapper, so make it first */
-	set_short_version(short_version, NULL);
+	write_version_file(NULL);
 
 	/* Select suitable configuration settings */
 	set_null_conf();
@@ -3131,12 +3085,12 @@ main(int argc, char *argv[])
 	setup_config();
 
 	/* Bootstrap template1 */
-	bootstrap_template1(short_version);
+	bootstrap_template1();
 
 	/*
 	 * Make the per-database PG_VERSION for template1 only after init'ing it
 	 */
-	set_short_version(short_version, "base/1");
+	write_version_file("base/1");
 
 	/* Create the stuff we don't need to use bootstrap mode for */
 
@@ -3158,6 +3112,8 @@ main(int argc, char *argv[])
 
 	setup_schema();
 
+	load_plpgsql();
+
 	vacuum_db();
 
 	make_template0();
